#include "imaget.h"
#include <string.h>

guiImplSizeResult guiImplSizeGif (FILE *in, unsigned long *wP, unsigned long *hP)
{
	unsigned char buffer[256];
	int ch;
	int colorTableEntries;
	if (fread(buffer, 1, 6, in) != 6) {
		return guiImplSizeWrongType;
	}
	if (memcmp(buffer, "GIF", 3)) {
		/* Not a GIF at all. */
		return guiImplSizeWrongType;
	}
	if (memcmp(buffer, "GIF87a", 6) && strncmp(buffer, "GIF89a", 6)) {
		/* A new version of the standard, so play nice
			and don't claim to understand it. */
		return guiImplSizeIndeterminate;
	}
	if (fread(buffer, 1, 7, in) != 7) {
		return guiImplSizeBadImage;
	}
	colorTableEntries = 2 << (buffer[4] & 0x07);
	if (buffer[4] & 0x80) {
		/* Skip the global color table if it's in the way */
		if (fseek(in, 3 * colorTableEntries, SEEK_CUR) != 0) {
			return guiImplSizeBadImage;
		}
	}

	while(1) {
		int extid, count;
		ch = getc(in);
		switch(ch) {
			case EOF:
			return guiImplSizeBadImage;
			case ';':
			/* No images?!? */
			return guiImplSizeBadImage;
			case '!':
			/* An extension. Skip data blocks. */
			extid = getc(in);
			if (extid == EOF) {
				return guiImplSizeBadImage;
			}
			do {
				count = getc(in);
				if (count == EOF) {
					return guiImplSizeBadImage;
				}
				if (count) {
					if (fseek(in, count, SEEK_CUR) != 0) {
						return guiImplSizeBadImage;
					}
				}
			} while (count);
			break;
			case ',':
			/* An image... finally. */
			if (fread(buffer, 1, 9, in) != 9) {
				return guiImplSizeBadImage;
			}
			/* YES! */
			*wP = (buffer[5] << 8) + buffer[4];	
			*hP = (buffer[7] << 8) + buffer[6];	
			return guiImplSizeSuccess;
			default:
			/* Bad file */
			return guiImplSizeBadImage;
		}
	}
}

guiImplSizeResult jpegFirstMarker (FILE *in)
{
	int ch1;
	int ch2;
	ch1 = getc(in);
	if (ch1 == EOF) {
		return guiImplSizeWrongType;
	}
	ch2 = getc(in);
	if (ch2 == EOF) { 
		return guiImplSizeWrongType;
	}
	if ((ch1 != 0xFF) && (ch2 != 0xD8)) {
		return guiImplSizeWrongType;
	}
	return guiImplSizeSuccess;
}
	
guiImplSizeResult jpegNextMarker (FILE *in, int *markerP)
{
	int c;
	int discarded = 0;

	c = getc(in);
	if (c == EOF) {
		return guiImplSizeBadImage;
	}
	while (c != 0xFF) {
		discarded++;
		c = getc(in);
		if (c == EOF) {
			return guiImplSizeBadImage;
		}
	}
	/* Get marker code byte, swallowing any duplicate FF bytes.  Extra FFs
	 * are legal as pad bytes, so don't count them in discarded.
	 */
	do {
		c = getc(in);
		if (c == EOF) {
			return guiImplSizeBadImage;
		}
	} while (c == 0xFF);
	if (discarded != 0) {
		/* Garbage data; not a good sign, but hopefully ignorable. */
	}
	*markerP = c;
	return guiImplSizeSuccess;
}

guiImplSizeResult guiImplSizeJpeg (FILE *in, unsigned long *wP, unsigned long *hP)
{
	guiImplSizeResult r;
	long length;
	int l1, l2;
	unsigned char buffer[4];
	r = jpegFirstMarker(in);
	if (r != guiImplSizeSuccess) {
		return r;
	}
	while(1) {
		int m;
		r = jpegNextMarker(in, &m);
		switch(m) {
			case 0xDA:
			case 0xD9:		
			/* Size can't be determined; leave it alone. */
			return guiImplSizeIndeterminate;
			break;
			case 0xC0:
			case 0xC1:
			case 0xC2:
			case 0xC3:
			case 0xC4:
			case 0xC5:
			case 0xC6:
			case 0xC7:
			case 0xC8:
			case 0xC9:
			case 0xCA:
			case 0xCB:
			case 0xCC:
			case 0xCD:
			case 0xCE:
			case 0xCF:
			if (fseek(in, 3, SEEK_CUR) != 0) {
				return guiImplSizeBadImage;
			}
			if (fread(buffer, 1, 4, in) != 4) {
				return guiImplSizeBadImage;
			}
			*hP = (buffer[0] << 8) + buffer[1];	
			*wP = (buffer[2] << 8) + buffer[3];	
			return guiImplSizeSuccess;
			default:
			/* A marker we don't know about. Skip it. */
			l1 = getc(in);
			if (l1 == EOF) {
				return guiImplSizeBadImage;
			}	
			l2 = getc(in);
			if (l2 == EOF) {
				return guiImplSizeBadImage;
			}	
			length = (l1 << 8) + l2;
			if (fseek(in, length - 2, SEEK_CUR) != 0) {
				return guiImplSizeBadImage;
			}
			break;
		}
	}
}

char pngSignature[] = { 137, 80, 78, 71, 13, 10, 26, 10 };

guiImplSizeResult guiImplSizePng (FILE *in, unsigned long *wP, unsigned long *hP)
{
	unsigned char buffer[256];
	if (fread(buffer, 1, 8, in) != 8) {
		return guiImplSizeWrongType;
	}
	if (memcmp(buffer, pngSignature, 8)) {
		return guiImplSizeWrongType;
	}	
	while (1) {
		unsigned long length;
		char chunkType[4];
		if (fread(buffer, 1, 4, in) != 4) {
			return guiImplSizeBadImage;
		}
		length = (buffer[0] << 24) + (buffer[1] << 16) + 
			(buffer[2] << 8) + (buffer[3]);
		if (fread(chunkType, 1, 4, in) != 4) {
			return guiImplSizeBadImage;
		}
		if (!memcmp(chunkType, "IEND", 4)) {
			/* Uh-uh. IHDR is required to appear before this. */
			return guiImplSizeBadImage;
		}
		if (!memcmp(chunkType, "IHDR", 4)) {
			/* Here we go */
			if (fread(buffer, 1, 4, in) != 4) {
				return guiImplSizeBadImage;
			}
			*wP = (buffer[0] << 24) + (buffer[1] << 16) + 
				(buffer[2] << 8) + buffer[3];
			if (fread(buffer, 1, 4, in) != 4) {
				return guiImplSizeBadImage;
			}
			*hP = (buffer[0] << 24) + (buffer[1] << 16) + 
				(buffer[2] << 8) + buffer[3];
			return guiImplSizeSuccess;	
		}
		/* Other chunks shouldn't precede IHDR, but try to
			humor the file and read on... */
		/* Advance past the data */
		if (fseek(in, length, SEEK_CUR) != 0) {
			return guiImplSizeBadImage;
		}
		/* Skip the CRC field */
		if (fread(buffer, 1, 4, in) != 4) {
			return guiImplSizeBadImage;
		}
	}
}

